#include <yulib/generic.h>
#include <yulib/memory.h>
#include <yulib/path.h>
#include <yulib/file.h>

static const BYTE PNG_SIGNATURE[] = { 0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A };
static const DWORD PNG_SIZEOF_CHUNKINFO = sizeof(DWORD) * 3;
static const LPCSTR PNG_FILEFILTER = "PNGt@C (*.png)\0*.png\0";

class CPngReader {
public:
	LPBYTE data;
	LPBYTE ptr;
	DWORD  size;

	CPngReader(void) : data(NULL), ptr(NULL), size(0) {}


	void Attach(LPBYTE data_, DWORD size_)
	{
		data = ptr = data_;
		size = size_;
	}

	void ResetPtr(void) { ptr = data; }

	DWORD PtrPos(void) { return (DWORD)(ptr - data); }
	DWORD Remaining(void) { return (DWORD)(size - PtrPos()); }

	bool IsPng(void)
	{
		return size >= sizeof(PNG_SIGNATURE) && !memcmp(data, PNG_SIGNATURE, sizeof(PNG_SIGNATURE));
	}

	DWORD GetChunkLength(void)
	{
		return yulib::Get4ByteBE(ptr);
	}

	bool NextChunk(void)
	{
		if(ptr == data) {
			if(!IsPng()) return false;
			ptr += sizeof(PNG_SIGNATURE);
		} else {
			if(Remaining() < PNG_SIZEOF_CHUNKINFO) return false;
			ptr += yulib::Get4ByteBE(ptr) + PNG_SIZEOF_CHUNKINFO;
		}

		return Remaining() >= yulib::Get4ByteBE(ptr) + PNG_SIZEOF_CHUNKINFO;
	}

	bool ChunkNameIs(LPCSTR s)
	{
		return *(DWORD*)(ptr + 4) == *(DWORD*)s;
	}

	void GetChunkName(LPSTR buf)
	{
		for(int i = 0; i < 4; i++) buf[i] = (char)ptr[i+4];
		buf[4] = '\0';
	}
};


void main(int argc, char** argv)
{
	yulib::CPath path;
	yulib::CMemory<BYTE> mem;
	CPngReader png_reader;

	// PNGt@Cǂݍ
	{
		if(argc >= 2) {
			path.Set(argv[1]);
		} else {
			puts("PNGt@CI");
			if(!path.OpenFileDialog(NULL, PNG_FILEFILTER, "PNGt@CI")) return;
		}
		puts(path);
		puts("");

		yulib::CFile file;
		if(!file.OpenExisting(path)) return;
		if(!mem.Alloc(file.Size(), false)) return;
		if(!file.Read(mem, mem.size)) return;
		png_reader.Attach(mem, mem.size);
		if(!png_reader.IsPng()) return;
	}

	// `Nꗗ\
	int chunk_num = 0;
	while(png_reader.NextChunk()) {
		char chunk_name[5];
		png_reader.GetChunkName(chunk_name);
		DWORD length = png_reader.GetChunkLength();
		printf("%d : %s %8d byte\r\n", chunk_num++, chunk_name, length);
	}
	png_reader.ResetPtr();
	puts("");

	// I
	int process = 0;
	printf("I(1:`N} 2:`No) : ");
	scanf_s("%d", &process);
	puts("");

	if(process == 1) {
		yulib::CFile file;
		yulib::CMemory<BYTE> insdata;
		// }t@CI
		puts("}t@CI");
		path.Clear();
		if(!path.OpenFileDialog(NULL, "ׂẴt@C (*.*)\0*.*\0", "}t@CI")) return;
		puts(path);
		puts("\r\n");

		// }t@Cǂݍ
		if(!file.OpenExisting(path)) return;
		DWORD filesize = file.Size();
		if(!insdata.Alloc(filesize + PNG_SIZEOF_CHUNKINFO, false)) return;
		if(!file.Read(insdata+8, filesize)) return;
		file.Close();

		yulib::Set4ByteBE(insdata, filesize);

		// }`N
		char chunk_name[5];
		printf("}`N : ");
		scanf_s("%4s", chunk_name, 5);
		for(int i = 0; i < 4; i++) insdata[i+4] = (BYTE)chunk_name[i];
		puts(chunk_name);
		puts("");

		png_reader.ResetPtr();
		while(!png_reader.ChunkNameIs("IEND") && png_reader.NextChunk());

		// CRCvZ
		DWORD crc = yulib::Crc32(insdata + 4, filesize+4);
		yulib::Set4ByteBE(insdata + insdata.size - 4, crc);

		// ۑt@C
		puts("ۑt@C");
		path.Clear();
		if(!path.SaveFileDialog(NULL, PNG_FILEFILTER, "ۑt@C")) return;
		if(!file.CreateAlways(path)) return;
		file.Write(png_reader.data, png_reader.PtrPos());
		file.Write(insdata, insdata.size);
		file.Write(png_reader.ptr, png_reader.Remaining());
		file.Close();
	} else if(process == 2) {
		// `N𒊏o
		int chunk_idx = 0;
		printf("o`NԍőI : ");
		scanf_s("%d", &chunk_idx);
		if(chunk_idx >= chunk_num) return;
		png_reader.ResetPtr();
		for(int i = 0; i <= chunk_idx; i++) png_reader.NextChunk();
		path.Clear();
		puts("o̓t@C");
		if(!path.SaveFileDialog(NULL, "ׂẴt@C (*.*)\0*.*\0", "o̓t@C")) return;
		yulib::OutputFile(path, png_reader.ptr + 8, png_reader.GetChunkLength());
	}
}
